import numpy as np


def locate_tumor(image, mask, label=1, padding=2):
    """
    Locate and extract tumor from CT image using mask
    Args:
        img: Numpy array. The whole image.
        mask: Numpy array. Same size as img, background is 0.
        label: Int. The value of mask foreground.
        padding: Int. Number of pixels padded on each side after extracting tumor.
    Returns:
        sub_img: Numpy array. The tumor area defined by mask.
        sub_mask: Numpy array. The subset of mask in the same position of sub_img.
    """
    top_margin = min(np.where(mask == label)[0])
    bottom_margin = max(np.where(mask == label)[0])
    front_margin = min(np.where(mask == label)[1])
    back_margin = max(np.where(mask == label)[1])
    left_margin = min(np.where(mask == label)[2])
    right_margin = max(np.where(mask == label)[2])
    sub_img = image[max(0, top_margin - padding):min(bottom_margin + padding + 1, image.shape[0]-1), 
                max(0, front_margin - padding):min(back_margin + padding + 1, image.shape[1]-1),
                max(0, left_margin - padding):min(right_margin + padding + 1, image.shape[2]-1)]
    sub_mask = mask[max(0, top_margin - padding):min(bottom_margin + padding + 1, mask.shape[0]-1), 
                max(0, front_margin - padding):min(back_margin + padding + 1, mask.shape[1]-1),
                max(0, left_margin - padding):min(right_margin + padding + 1, mask.shape[2]-1)]
    return sub_img, sub_mask




import sys
sys.path.append('.')

from .load import load_image, load_seg
from .locate import locate_tumor
from .get_projection import get_model, get_pca
from .inference import model_inference, pca_trans
from .visualize import visualize_reference, visualize
from .calPPS import calPPS
from .utils import cal_tumor_size, get_largest_slice





import numpy as np
import pickle
from rpy2.robjects import r
from rpy2.robjects.packages import importr
from rpy2.robjects import pandas2ri

def calPPS(infer_data, reference_data_path="./data/ref_data.json", traj_obj_path="./data/traj_obj.rds"):
    """
    Calculate the Pseudo-Progression Score (PPS) for the inference data.
    Args:
        infer_data (Numpy array): The PCA embedding(s) of inference data, should be in shape (N, 2)
        reference_data_path (str, optional): Path to the reference data. Defaults to "./data/ref_data.json".
        traj_obj_path (str, optional): Path to the trajecory object. . Defaults to "./data/traj_obj.rds".

    Returns:
        infer_pred_ps: The inferenced PPS which quantifies the progression status of early-stage LUAD.
    """
    pandas2ri.numpy2ri.activate()
    slingshot = importr("slingshot")

    with open(reference_data_path, 'rb') as f:
        ref_data = pickle.load(f)
    train_data = ref_data["pca_embed"]
    traj_obj = r.readRDS(traj_obj_path)

    infer_len = infer_data.shape[0]
    infer_data_aug = r.rbind(infer_data, train_data)
    infer_pred_aug = slingshot.predict(slingshot.as_SlingshotDataSet(traj_obj), infer_data_aug)
    infer_pred_ps_aug = slingshot.slingPseudotime(infer_pred_aug).flatten()
    infer_pred_ps = infer_pred_ps_aug[:infer_len]

    return infer_pred_ps



import matplotlib.pyplot as plt
import numpy as np
import pickle


def visualize_reference(show_curve=True, reference_data_path="./data/ref_data.json", curve_data_path="./data/curve_data.npy", transparency=1.0):
    """
    Visualize the distribution of reference data. Users can choose whether to show the radiomic trajectory.
    Args:
        show_curve (bool, optional): Whether to show the trajectory together with data points. Defaults to True.
        reference_data_path (str, optional): Path to the reference data. Defaults to "./data/ref_data.npy".
        curve_data_path (str, optional): Path to the curve points data. Defaults to "./data/curve_data.npy".
        transparency (float, optional): The degree of transparency of reference data points. Defaults to 1.0.
    Returns:
        fig: Generated figure.
    """
    with open(reference_data_path, 'rb') as f:
        ref_data = pickle.load(f)
    pca_embed = ref_data["pca_embed"]
    labels = ref_data["labels"]

    fig = plt.figure(figsize=(10,3))
    c0 = plt.scatter(pca_embed[labels==0,0], pca_embed[labels==0,1], c='blue', label='AIS', s=10, alpha=transparency)
    c1 = plt.scatter(pca_embed[labels==1,0], pca_embed[labels==1,1], c='lime', label='MIA', s=10, alpha=transparency)
    c2 = plt.scatter(pca_embed[labels==2,0], pca_embed[labels==2,1], c='olive', label='AIC-I', s=10, alpha=transparency)
    c3 = plt.scatter(pca_embed[labels==3,0], pca_embed[labels==3,1], c='orange', label='IAC-II', s=10, alpha=transparency)
    c4 = plt.scatter(pca_embed[labels==4,0], pca_embed[labels==4,1], c='red', label='IAC-III', s=10, alpha=transparency)
    plt.legend()
    plt.xlabel("PCA_1")
    plt.ylabel("PCA_2")
    plt.xlim([-1.6,2])
    plt.ylim([-0.6, 0.6])

    if show_curve:
        curve = np.load(curve_data_path)
        plt.plot(curve[:,0], curve[:,1], c='black', linewidth=0.8)
    
    return fig


def visualize(pca_embed, show_reference=True, reference_data_path="./data/ref_data.json", curve_data_path="./data/curve_data.npy", transparency=0.2):
    """
    Visualize the distribution of query tumor. Users can choose whether to show the refernce data.
    Args:
        pca_embed (_type_): The PCA embedding of inference data, should be in shape (N, 2)
        show_reference (bool, optional): Whether to show reference data and trajectory. Defaults to True.
        transparency (float, optional): The degree of transparency of reference data points. Defaults to 0.2.
    Returns:
        fig: Generated figure.
    """
    if show_reference:
        fig = visualize_reference(reference_data_path=reference_data_path, curve_data_path=curve_data_path, transparency=transparency)
        plt.scatter(pca_embed[:,0], pca_embed[:,1], s=10, c='purple')
    else:
        fig = plt.figure(figsize=(10,3))
        plt.scatter(pca_embed[:,0], pca_embed[:,1], s=10, c='purple')
        plt.xlim([-1.6,2])
        plt.ylim([-0.6, 0.6])
        plt.xlabel("PCA_1")
        plt.ylabel("PCA_2")
    
    return fig






import monai
import torch
import torch.nn as nn
import pickle


class Model(nn.Module):
    def __init__(self, net, embed_dim=32):
        super(Model, self).__init__()
        self.net = net
        self.flatten = nn.Flatten(start_dim=1, end_dim=-1)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(0.5)
        self.fc1 = nn.Linear(2048, 128, bias=True)
        self.fc2 = nn.Linear(129, embed_dim, bias=True)
        
    def forward(self, x, add_variable):
        x = self.net(x)
        x = self.flatten(x)
        x = torch.cat((self.dropout(self.relu(self.fc1(x))), add_variable.unsqueeze(1)),1)
        x = self.fc2(x)
        return x


def get_model(wpath="./data/cnn_proj_weights.pkl"):
    """
    Args:
        wpath (str): path to the saved model weights file.
    Returns:
        model: a CNN model with pre-trained weights loaded. (In CPU)
    """
    temp_model = monai.networks.nets.Classifier(in_shape=(1,64,64,64), classes=5, channels=(8,8,4), strides=(2,2,2))
    net = temp_model.net
    model = Model(net)

    weights = torch.load(wpath)
    model.load_state_dict(weights)

    return model


def get_pca(ppath="./data/pca_trans.sav"):
    """
    Args:
        ppath (str): path to the saved PCA transform function.
    Returns:
        pca_trans: a pre-trained PCA transform function
    """
    return pickle.load(open(ppath, 'rb'))




import numpy as np
import SimpleITK as sitk
import six
from radiomics import featureextractor

def read_dcm_series(dcm_dir):
    """
    Args:
        dcm_dir: Str. Path to dicom series directory
    Returns:
        sitk_image: SimpleITK object of 3D CT volume.
    """
    reader = sitk.ImageSeriesReader()
    series_file_names = reader.GetGDCMSeriesFileNames(dcm_dir)
    reader.SetFileNames(series_file_names)
    sitk_image = reader.Execute()

    return sitk_image


def get_largest_slice(img3d, mask3d):
    """
    Get the slice with largest tumor area
    Args:
        img3d: Numpy array. The whole CT volume (3D)
        mask3d: Numpy array. Same size as img3d, binary mask with tumor area set as 1, background as 0
    Returns:
        img: Numpy array. The 2D image slice with largest tumor area
        mask: Numpy array. The subset of mask in the same position of sub_img
    """
    area = np.sum(mask3d == 1, axis=(1, 2))
    area_index = np.argsort(area)[-1]
    img = img3d[area_index, :, :]
    mask = mask3d[area_index, :, :]

    return img, mask


def cal_tumor_size(mask):
    """
    Args:
        mask: Numpy array. The mask of tumor with the same shape of image. Can be either the single mask or a list of masks.
    Returns:
        t_size: Int. The size of tumor.
    """
    mask = np.array(mask)
    if len(mask.shape)==4:
        return np.sum(mask, axis=(1,2,3))
    elif len(mask.shape)==3:
        return np.sum(mask)
    else:
        raise Exception("Mask size should be either single mask or a list of masks")



import os
import SimpleITK as sitk
from .utils import read_dcm_series


def load_image(path):
    """
    Load CT image volume
    Args:
        path: Str. Path to the .nii(.gz) file or dicom series directory
    Returns:
        image: Numpy array. The 3D CT volume.
    """
    if os.path.isdir(path):
        sitk_image = read_dcm_series(path)
    else:
        sitk_image = sitk.ReadImage(path)
    image = sitk.GetArrayFromImage(sitk_image)

    return image


def load_seg(path):
    """
    Load segmentation mask
    Args:
        path: Str. Path to the .nii or .dcm mask.
    Returns:
        seg: Numpy array. The mask of ROI with the same shape of image.
    """
    if path.endswith(".dcm"):
        # RTStruct dcm file sometimes cannot be loaded by SimpleITK
        ds = dicom.read_file(path)
        seg = ds.pixel_array
    else:
        sitk_seg = sitk.ReadImage(path)
        seg = sitk.GetArrayFromImage(sitk_seg)

    return seg


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="radiotrace",
    version="0.2.0",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "numpy>=1.17.0",
        "matplotlib>=3.1.3",
        "scikit-learn>=0.23.2",
        "scipy>=1.5.2",
        "pydicom>=1.4.1",
        "SimpleITK>=1.2.4",
        "torch>=1.10.0",
        "monai>=0.8.0",
        "rpy2>=3.5.6"
    ],
    author="Jiaqi Li",
    author_email="li-jq18@mails.tsinghua.edu.cn",
    description="package for quantify early-stage LUAD progression from CT image",
    license="MIT",
    url="https://github.com/LiJiaqi96/radiotrace",
    long_description=long_description,
    long_description_content_type="text/markdown"
)



import torch
from torch.utils.data import Dataset, DataLoader
import monai
from monai.transforms import AddChannel, Resize, ScaleIntensity, EnsureType, Compose
from sklearn.decomposition import PCA


class earlyLUADDataset(Dataset):
    def __init__(self, tumors, sizes, transform):
        self.tumors = tumors
        self.sizes = sizes
        self.transform = transform
    
    def __len__(self):
        return len(self.sizes)
    
    def __getitem__(self, index):
        tumor = self.tumors[index]
        size = torch.scalar_tensor(self.sizes[index]/16000.)
        trans_tumor = self.transform(tumor)
        # tumor_tensor = torch.tensor(trans_tumor, dtype=torch.float)
        return trans_tumor, size


def model_inference(images, sizes, model, batch_size=1, unisize=(64,64,64), use_GPU=False):
    """
    Args:
        images (Numpy array or Torch tensor): The input tumor(s), should be in shape (N C Z H W).
        sizes (List of int): The corresponding tumor sizes, should be in shape (N 1).
        model (nn.Module): The loaded pre-trained CNN model.
        batch_size (Int): Batch size for model inference. Defaults to 1.
        unisize (tuple, optional): The uniformed size of input image. Defaults to (64,64,64).
    Returns:
        outputs (Numpy array): The embedding vector of the tumor, should be in shape (B 32)
    """
    trans_func = Compose([AddChannel(), Resize(spatial_size=unisize), ScaleIntensity(), EnsureType()])
    infer_data = earlyLUADDataset(images, sizes, trans_func)
    infer_loader = DataLoader(infer_data, batch_size=batch_size, num_workers=0)

    if use_GPU:
        model = model.cuda()
        model.eval()
        with torch.no_grad():
            preds = []
            for inputs, sizes in infer_loader:
                inputs = inputs.cuda()
                sizes = sizes.cuda()
                output = model(inputs, sizes)
                preds.append(output)
            preds = torch.concat(preds).cpu().data.numpy()
    else:
        model.eval()
        with torch.no_grad():
            preds = []
            for inputs, sizes in infer_loader:
                output = model(inputs, sizes)
                preds.append(output)
            preds = torch.concat(preds).numpy()
    
    return preds


def pca_trans(embed, pca):
    """
    Args:
        embed (Numpy array): The embedding vectors of tumors, should in shape (N, 32), N is the inference sample size and 32 is the embedding dimension.
    Returns:
        pca_embed (Numpy array): The coordinates of embedding vectors on the PCA space.
    """
    return pca.transform(embed)

